#!/usr/bin/python3

import os, sys
import json, pickle
import shutil
from uuid import uuid4
from datetime import datetime, timezone
import base64
from hashlib import sha1
import argparse


HOMEDIR = f"/home/{os.getlogin()}"
HOST = "ctrl-c.club"

class IrisError(Exception):
    ...


class Messages:
    read = []

    topics = []
    messages = []

    topics_latest = []
    messages_latest = []

    def __init__(self, corpus=None):
        self.__update()

    def __update(self):
        """update state of `corpus`, `read`, `topics` and `messages` lists"""

        corpus = []
        userlist = os.listdir('/home/')

        for user in userlist:
            try:
                if '.iris.messages' in os.listdir(f'/home/{user}/'):
                    with open(f'/home/{user}/.iris.messages', 'r') as messages:
                        userdata = json.loads(messages.read())
                        if userdata != []:
                            corpus += userdata

            except (PermissionError, NotADirectoryError):
                pass

        # sort corpus in ascending order by timestamp
        corpus = sorted(corpus, key=lambda k: k['data']['timestamp'])

        # get hashes of read messages from the user's `.irispy.messages.read`
        with open(f"{HOMEDIR}/.irispy.messages.read", "rb") as f:
            self.read = pickle.load(f)

        # topics and messages lists
        for item in corpus:
            if item["data"]["parent"] == None:
                self.topics.append(item)
            else:
                self.messages.append(item)

        self.topics_latest_version = [item for item in self.topics if item["edit_hash"] == None]
        self.messages_latest_version = [item for item in self.messages if item["edit_hash"] == None]

        for i, topic in enumerate(self.topics_latest_version): # topic numbers
            topic["index"] = i+1

    def unread_topics(self):
        """returns a list of topics with unread messages"""
        output = []
        
        for topic in self.topics_latest_version:
            topic_hash = topic["hash"]

            unread_reply_count = 0
            reply_count = 0
            for message in self.messages_latest_version:
                if message["data"]["parent"] == topic_hash:
                    if message["hash"] not in self.read:
                        unread_reply_count += 1
                    reply_count += 1

            if reply_count == 0 or unread_reply_count > 0:
                output.append(topic)

        return output

    def __get_last_edit(self, topic_hash):
        for message in self.topics:
            if message["edit_hash"] == topic_hash:
                return message["hash"]
        return None

    def __get_editchain(self, topic_hash):
        editchain = [topic_hash]

        while self.__get_last_edit(editchain[-1]):
            editchain.append(self.__get_last_edit(editchain[-1]))

        return editchain

    def replies_to(self, topic_hash):
        all_versions = self.__get_editchain(topic_hash)

        output = []
        for msg in self.messages_latest_version:
            if msg["data"]["parent"] in all_versions:
                output.append(msg)
        
        return output
    
    def mark_as_read(self, messages):
        for msg in messages:
            if msg["hash"] not in self.read:
                self.read.append(msg["hash"])

        with open(f"{HOMEDIR}/.irispy.messages.read", "wb") as f:
            pickle.dump(self.read, f)
    
    def mark_all_read(self):
        self.mark_as_read(self.messages_latest_version)

    def get_hash_from_topic_no(self, num):
        return self.topics_latest_version[num - 1]["hash"]

    def __write_message(self) -> str:
        """launch an editor to write a message and return message string"""

        tmp_filename = "irispy_tmp_" + uuid4().hex
        tmp_filepath = f"{HOMEDIR}/{tmp_filename}"

        content = ""
        with open(tmp_filepath, 'w') as f:
            editor = os.environ['EDITOR']
            if editor == None:
                editor = "nano -t"
            os.system(f"{editor} {tmp_filepath}")

        with open(tmp_filepath, 'r') as f:
            content = f.read()

        # delete temporary file
        os.system(f"rm {tmp_filepath}")

        return content

    def create_post(self, parent=None):
        author = f"{os.getlogin()}@{HOST}"
        content = self.__write_message()

        if len(content) == 0:
            raise IrisError("post cannot be empty")

        # GMT Timestamp
        timestamp = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")

        # Message SHA1 hash from schema defined in iris reference implementation
        hash_source = json.dumps({
                "author": author,
                "parent": parent,
                "timestamp": timestamp,
                "message": content
                }, separators=(',', ':')
            ).encode()

        message_hash = base64.b64encode(sha1(hash_source).digest())

        # create message object, append it to .iris.messages
        message_obj = {
                "hash": message_hash.decode('utf-8') + '\n',
                "edit_hash": None,
                "is_deleted": None,
                "data": {
                    "author": author,
                    "parent": parent,
                    "timestamp": timestamp,
                    "message": content
                }
            }
        
        # confirmation
        if input("save message? [Y/n] ") in ('N','n'):
            print("canceled")
            sys.exit(0)

        # updating .iris.messages
        user_messages = []
        with open(f"{HOMEDIR}/.iris.messages", "r") as f:
            user_messages = json.loads(f.read())

        user_messages.append(message_obj)

        with open(f"{HOMEDIR}/.iris.messages", "w") as f:
            f.write(json.dumps(user_messages))

        print("message saved")


class Display:
    def print_as_feed(messages):
        COLS = shutil.get_terminal_size((80, 20))[0]
        max_namelen = max([len(msg["data"]["author"].split('@')[0]) for msg in messages])
        tstamp_len = 10

        max_numlen = len(str(max([msg["index"] for msg in messages])))

        for msg in messages:
            author = msg["data"]["author"].split('@')[0]
            content_part = msg["data"]["message"].split('\n')[0].replace('\n','').replace('\t',' ')
            tstamp = msg["data"]["timestamp"].split('T')[0]
            topic_no = msg["index"]

            try:
                if msg["is_deleted"]:
                    #print(f" {topic_no}{' '*(max_numlen-len(str(topic_no)))} \x1b[33m{author}{' '*(max_namelen-len(author))}\033[00m \x1b[37m{tstamp}\033[00m DELETED BY AUTHOR")
                    continue
            except KeyError:
                print("<malformed message object>")
                continue

            print(f" {topic_no}{' '*(max_numlen-len(str(topic_no)))} \x1b[33m{author}{' '*(max_namelen-len(author))}\033[00m \033[96m{tstamp}\033[00m {content_part[:(COLS-max_namelen-tstamp_len-max_numlen-5)]}")

    def print_message(msg):
        COLS = shutil.get_terminal_size((80, 20))[0]
        text = msg["data"]["message"]
        tstamp = msg["data"]["timestamp"].split('T')[0]
        author = msg["data"]["author"].split('@')[0]

        space = COLS - (len(author) + len(tstamp) + 1) - 3

        print(f" \x1b[33m{author}\033[00m \x1b[37m{tstamp}\033[00m {'-'*space}")
        print(text)
        if text[-1] != '\n':
            print()

    def view_topic(topic_no, messages):
        COLS = shutil.get_terminal_size((80, 20))[0]

        topic_no = int(topic_no)
        
        requested_topic = None
        for topic in messages.topics_latest_version:
            if int(topic["index"]) == topic_no:
                requested_topic = topic
                break

        if topic["is_deleted"] != None:
            print("This topic has been deleted by its author")
            sys.exit(0)

        replies = messages.replies_to(requested_topic["hash"])
        messages.mark_as_read(replies)

        Display.print_message(requested_topic)
        print(f" {'='*(COLS-2)} ")

        for msg in replies:
            Display.print_message(msg)


if __name__ == "__main__":
    # first-run stuff
    ## creating `.irispy.messages.read` if it doesn't exist
    if ".irispy.messages.read" not in os.listdir(HOMEDIR):
        read_messages = []
        with open(f"{HOMEDIR}/.iris.messages.read", 'r') as f:
            read_messages = json.loads(f.read())
        with open(f"{HOMEDIR}/.irispy.messages.read", "wb") as f:
            pickle.dump(read_messages, f)

    # parsing arguments
    messages = Messages()

    parser = argparse.ArgumentParser()

    parser.add_argument("-u", "--unread", help="list topics with unread messages", action="store_true")
    parser.add_argument("-t", "--topics", help="list all topics", action="store_true")

    parser.add_argument("-v", "--view", help="view topic N", metavar='N', nargs=1, type=lambda x: int(x) if int(x) > 0 else False)
    parser.add_argument("-r", "--reply", help="reply to topic N", metavar='N', nargs=1, type=lambda x: int(x) if int(x) > 0 else False) 
    parser.add_argument("-p", "--post", help="create new topic", action="store_true")

    parser.add_argument("--mark-all-read", help="mark all topics as read", action="store_true")

    args = parser.parse_args()

    if args.unread:
        topics = messages.unread_topics()
        if len(topics) == 0:
            print("you're all caught up!")
        else:
            Display.print_as_feed(messages.unread_topics())
        sys.exit(0)

    if args.topics:
        Display.print_as_feed(messages.topics_latest_version)
        sys.exit(0)

    if args.view:
        topic_no = args.view[0]
        if not topic_no:
            raise IrisError("invalid topic number")
        
        Display.view_topic(topic_no, messages)
        
        sys.exit(0)

    if args.mark_all_read:
        messages.mark_all_read()
        print("all topics marked read")
        sys.exit(0)

    if args.post:
        messages.create_post()
        sys.exit(0)

    if args.reply:
        topic_no = args.reply[0]
        if not topic_no:
            raise IrisError("invalid topic number")

        messages.create_post(messages.get_hash_from_topic_no(topic_no))

        sys.exit(0)

    # default behaviour - show unread topics
    Display.print_as_feed(messages.unread_topics())
        
